package cn.dankal.test.common.service.aspect;

import com.alibaba.fastjson.JSON;
import lombok.Data;
import lombok.experimental.Accessors;
import org.aspectj.lang.JoinPoint;
import org.aspectj.lang.ProceedingJoinPoint;
import org.aspectj.lang.annotation.AfterThrowing;
import org.aspectj.lang.annotation.Around;
import org.aspectj.lang.annotation.Aspect;
import org.aspectj.lang.annotation.Pointcut;
import org.aspectj.lang.reflect.MethodSignature;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.stereotype.Component;
import org.springframework.util.Assert;
import org.springframework.web.context.request.RequestContextHolder;
import org.springframework.web.context.request.ServletRequestAttributes;
import org.springframework.web.multipart.MultipartFile;

import javax.servlet.http.HttpServletRequest;
import java.io.Serializable;
import java.util.HashMap;
import java.util.Map;

/**
 * {@link Aspect,Component}
 * 请求日志切面类
 *
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2020/10/19
 * @see DoRequestLog
 */
@Aspect
@Component
public class RequestLogAspect {

    /**
     * LOGGER
     */
    private final static Logger LOGGER = LoggerFactory.getLogger(RequestLogAspect.class);

    /**
     * 切入点表达式
     */
    @Pointcut("@within(cn.dankal.test.common.service.aspect.DoRequestLog)")
    public void pointcut() {
    }

    /**
     * Spring 环绕通知: 打印请求信息
     *
     * @param proceedingJoinPoint ProceedingJoinPoint
     * @return 目标方法返回值
     * @throws Throwable 目标方法执行异常
     * @see ProceedingJoinPoint,RequestContextHolder
     */
    @Around("pointcut()")
    public Object doAround2RequestLog(ProceedingJoinPoint proceedingJoinPoint) throws Throwable {

        long startTime = System.currentTimeMillis();

        // HttpServletRequest
        ServletRequestAttributes attributes = (ServletRequestAttributes) RequestContextHolder.getRequestAttributes();
        Assert.notNull(attributes, "doAround2RequestLog  --> 获取 ServletRequestAttributes 失败!!!");
        HttpServletRequest request = attributes.getRequest();

        Object result = proceedingJoinPoint.proceed();

        // HttpServletRequest --> RequestInfo
        RequestInfo requestInfo = new RequestInfo()
                .setIp(request.getRemoteAddr())
                .setUrl(request.getRequestURL().toString())
                .setHttpMethod(request.getMethod())
                .setClassMethod(String.format("%s.%s", proceedingJoinPoint.getSignature().getDeclaringTypeName(),
                        proceedingJoinPoint.getSignature().getName()))

                .setRequestParams(getRequestParamsByProceedingJoinPoint(proceedingJoinPoint))
                .setResult(result)
                .setTimeCost(System.currentTimeMillis() - startTime);
        // 参考 <<阿里巴巴Java开发手册>>(嵩山版) 日志输出不应直接使用 JSON 工具将其转化成 String see: 二 异常日志  ==>  (三) 日志规约: 10
        // 先谈谈为啥需要选择 序列化成 JSON String: 方便我们借助 JSON 工具更优雅的查看日志记录信息
        // 为此 加入 try-catch 避免 日志序列化导致服务不可用
        try {
            LOGGER.info("Request-Info : {}", JSON.toJSONString(requestInfo));
        } catch (Exception e) {
            // 序列化失败
            // 使用 toString  代替
            LOGGER.info("use toString  method --> Request-Info : {}", requestInfo.toString());
        }
        return result;
    }


    /**
     * Spring 异常返回通知: 打印错误日志信息
     *
     * @param joinPoint JoinPoint
     * @param throwable Throwable
     * @see JoinPoint,Throwable
     */
    @AfterThrowing(pointcut = "pointcut()", throwing = "throwable")
    public void doAfterThrow2RequestLog(JoinPoint joinPoint, Throwable throwable) {
        // HttpServletRequest
        ServletRequestAttributes attributes = (ServletRequestAttributes) RequestContextHolder.getRequestAttributes();
        Assert.notNull(attributes, "doAfterThrow2RequestLog --> 获取 ServletRequestAttributes 失败!!!");
        HttpServletRequest request = attributes.getRequest();

        // HttpServletRequest --> RequestErrorInfo
        RequestErrorInfo requestErrorInfo = new RequestErrorInfo()
                .setIp(request.getRemoteAddr())
                .setUrl(request.getRequestURL().toString())
                .setHttpMethod(request.getMethod())
                .setClassMethod(String.format("%s.%s", joinPoint.getSignature().getDeclaringTypeName(),
                        joinPoint.getSignature().getName()))
                .setRequestParams(getRequestParamsByJoinPoint(joinPoint))
                .setThrowable(throwable);
        // 参考 <<阿里巴巴Java开发手册>>(嵩山版) 日志输出不应直接使用 JSON 工具将其转化成 String see: 二 异常日志  ==>  (三) 日志规约: 10
        // 先谈谈为啥需要选择 序列化成 JSON String: 方便我们借助 JSON 工具更优雅的查看日志记录信息
        // 为此 加入 try-catch 避免 日志序列化导致服务不可用
        try {
            LOGGER.error("Error-Request-Info  : {}", JSON.toJSONString(requestErrorInfo));
        } catch (Exception e) {
            // 序列化失败
            // 使用 toString  代替
            LOGGER.info("use toString  method --> Request-Info : {}", requestErrorInfo.toString());
        }
    }

    /**
     * 通过 ProceedingJoinPoint 获取请求参数信息
     *
     * @param proceedingJoinPoint ProceedingJoinPoint
     * @return 封装了请求信息的 Map<String, Object>
     * @see ProceedingJoinPoint
     */
    private Map<String, Object> getRequestParamsByProceedingJoinPoint(ProceedingJoinPoint proceedingJoinPoint) {
        //参数名
        String[] paramNames = ((MethodSignature) proceedingJoinPoint.getSignature()).getParameterNames();
        //参数值
        Object[] paramValues = proceedingJoinPoint.getArgs();

        return buildRequestParamInfo(paramNames, paramValues);
    }

    /**
     * 通过  JoinPoint 后期请求参数信息
     *
     * @param joinPoint JoinPoint
     * @return 装了请求信息的 Map<String, Object>
     * @see JoinPoint
     */
    private Map<String, Object> getRequestParamsByJoinPoint(JoinPoint joinPoint) {
        //参数名
        String[] paramNames = ((MethodSignature) joinPoint.getSignature()).getParameterNames();
        //参数值
        Object[] paramValues = joinPoint.getArgs();

        return buildRequestParamInfo(paramNames, paramValues);
    }

    /**
     * 构建 请求参数信息
     *
     * @param paramNames  请求参数名称数组
     * @param paramValues 请求参数值数组
     * @return Map<String, Object>
     */
    private Map<String, Object> buildRequestParamInfo(String[] paramNames, Object[] paramValues) {
        Map<String, Object> requestParams = new HashMap<>(1 << 5);
        for (int i = 0; i < paramNames.length; i++) {
            Object value = paramValues[i];
            //如果是文件对象
            if (value instanceof MultipartFile) {
                MultipartFile file = (MultipartFile) value;
                //获取文件名
                value = file.getOriginalFilename();
            }
            requestParams.put(paramNames[i], value);
        }
        return requestParams;
    }


    /**
     * 请求信息记录
     */
    @Data
    @Accessors(chain = true)
    static class RequestInfo implements Serializable {
        private static final long serialVersionUID = -8109854803896597581L;
        /**
         * 请求 ip
         */
        private String ip;
        /**
         * 请求 url
         */
        private String url;
        /**
         * 请求方式
         */
        private String httpMethod;
        /**
         * 处理器方法
         */
        private String classMethod;
        /**
         * 请求参数
         */
        private Object requestParams;
        /**
         * 响应结果
         */
        private Object result;
        /**
         * 请求处理消耗时间
         */
        private Long timeCost;
    }


    /**
     * 请求错误信息记录
     */
    @Data
    @Accessors(chain = true)
    static class RequestErrorInfo implements Serializable {
        private static final long serialVersionUID = 8966927431771824888L;
        /**
         * 请求 ip
         */
        private String ip;
        /**
         * 请求 url
         */
        private String url;
        /**
         * 请求方式
         */
        private String httpMethod;
        /**
         * 处理器方法
         */
        private String classMethod;
        /**
         * 请求参数
         */
        private Object requestParams;
        /**
         * 异常信息
         */
        private Throwable throwable;
    }
}

